/********************************************************************
 *                                                                  *
 * HAV Library, Copyright 1994-2004 Gromada.com                     *
 *                                                                  *
 * THIS FILE IS PART OF THE "HAV Library" SOURCE CODE.              *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A BSD-STYLE SOURCE LICENSE INCLUDED WITH THIS SOURCE *
 * IN 'LICENSE.TXT'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.   *
 *                                                                  *
 ********************************************************************
 *                                                                  *
 * codecs.c - video data compressing / decompressing routines       *
 *                                                                  *
 ********************************************************************/


#include  <stdlib.h>
#include  <string.h>

#include  "hav.h"


// local consts
#define  METHOD_UNC  0  // uncompresed
#define  METHOD_PFR  1  // previous frame
#define  METHOD_RLE  2  // run-length, RLE
#define  METHOD_PLN  3  // previous line
#define  METHOD_SQR  4  // square block (BEST codec)
#define  METHOD_IDX  5  // table square (BEST codec)
//
#define  MAXUNC           1  // max. bytes per UNC packet
#define  MAXPLN_FAST     65  // max. bytes per PLN packet (FAST codec)
#define  MAXPLN_BEST    129  // max. bytes per PLN packet (BEST codec)
#define  MAXRLE         767  // max. pixels per RLE packet
#define  MAXPFR       49089  // max. bytes per PFR packet
#define  MAXSQR         767  // max. pixels per RLE packet (BEST codec)
#define  MAXIDX        1024  // max. bytes per PFR packet (BEST codec)
// size in pixels of square block used by BEST codec
#define  SQR_HALF_XRES  64
#define  SQR_YRES       12


typedef struct {
  Uchar   *bufferP;  // pointer to compressed data
  Int32   bitsFree;
  Uint32  bitsValue;
  Int32   bitsShift;
} HavBits;


// start of code
static void  memmove_safe_ (
  Uchar   *dest,
	Uchar   *src,
	Uint32  len)
// copy bytes of memory in forward direction, memory may overlap, memcpy and memmove doesn't comply to these rules
{
  for (; len > 0; len--)
    *(dest++) = *(src++);
}


static void  bits_init_ (
  HavBits  *bitsP,
  Pointer  cmpsDataP)
{
  bitsP->bitsFree = 8;
  bitsP->bufferP = cmpsDataP;
}


static void  bits_out_ (
  HavBits  *bitsP,
  Uint32   value,
  Int32    numBits)
{
  bitsP->bitsFree -= numBits;
  if (bitsP->bitsFree < 0) {
    *bitsP->bufferP = (*bitsP->bufferP << (bitsP->bitsFree + numBits)) | (value >> abs (bitsP->bitsFree));
    bitsP->bitsFree += 8;
    while (bitsP->bitsFree < 0) {
      *(++bitsP->bufferP) = value >> abs (bitsP->bitsFree);
      bitsP->bitsFree += 8;
    }
    *(++bitsP->bufferP) = value;
  }
  else
    *bitsP->bufferP = (*bitsP->bufferP << numBits) | value;
}


static void  bits_done_ (
  HavBits  *bitsP)
{
  if (bitsP->bitsFree != 8) {
    *bitsP->bufferP <<= bitsP->bitsFree;
    bitsP->bufferP++;
  }
}


static void  length_out_ (
  HavBits  *bitsP,
  Uint32   len)
{                                                   //      RLE           PFR
  if (len <= 3) bits_out_ (bitsP, len, 2);          //     2..3     66....193
  else if (len <= 7) bits_out_ (bitsP, len, 4);     //     4..7    194....449
  else if (len <= 15) bits_out_ (bitsP, len, 6);    //    8..15    450....961
  else if (len <= 31) bits_out_ (bitsP, len, 8);    //   16..31    962...1985
  else if (len <= 63) bits_out_ (bitsP, len, 10);   //   32..63   1986...4033
  else if (len <= 127) bits_out_ (bitsP, len, 12);  //  64..127   4034...8129
  else if (len <= 255) bits_out_ (bitsP, len, 14);  // 128..255   8130..16321
  else bits_out_ (bitsP, len-256, 16);              // 256..767  16322..49089
}


static Uint32  hav_encode_fast (
  Uchar   *cmpsDataP,
	Uchar   *imgDataP,
	Uchar   *prevImgDataP,
	Uint32  resX,
	Uint32  resY,
	Uint32  bpp)
{
  Bool     usePRFmethod = !!prevImgDataP;
  Uint32   imgPos=0, tmpPos, tmpPosPL, cnt, tmpCnt, method;
  Uint32   endData, resZ, lineSize, imgSize;
  Uchar    *pByte, *cmpsMax;
  Uchar    color1, color2, color3;
  Ushort   *pWord;
  HavBits  bits;

  bits_init_ (&bits, cmpsDataP);
  resZ = (bpp + 7) / 8;
  lineSize = resX * resZ;
  imgSize = lineSize * resY;
  cmpsMax = cmpsDataP + (Uint32) (imgSize * 0.95);
  do {
    cnt = MAXUNC;
    method = METHOD_UNC;
    if (usePRFmethod) {
      tmpPos = imgPos;
      if ((endData = imgPos+MAXPFR) > imgSize)
        endData = imgSize;
      while (tmpPos < endData && imgDataP[tmpPos] == prevImgDataP[tmpPos]) tmpPos++;
      if ((tmpCnt = tmpPos-imgPos) > cnt) {
        cnt = tmpCnt;
        method = METHOD_PFR;
      }
    }
    if (cnt < MAXRLE) {
      tmpCnt = (imgSize-imgPos) / resZ;
      if ((endData = MAXRLE) > tmpCnt)
        endData = tmpCnt;
      tmpCnt = 1;
      switch (resZ) {
        case 1:
          pByte = &imgDataP[imgPos];
          while (tmpCnt < endData && pByte[0] == pByte[tmpCnt]) tmpCnt++;
          break;
        case 2:
          pWord = (Ushort *) &imgDataP[imgPos];
          while (tmpCnt < endData && pWord[0] == pWord[tmpCnt]) tmpCnt++;
          break;
        case 3:
          pByte = &imgDataP[imgPos];
          color1 = *(pByte++);
          color2 = *(pByte++);
          color3 = *(pByte++);
          while (tmpCnt < endData && color1 == *(pByte++)
            && color2 == *(pByte++) && color3 == *(pByte++)) tmpCnt++;
          break;
      }
      if (tmpCnt > 1 && tmpCnt*resZ > cnt) {
        cnt = tmpCnt * resZ;
        method = METHOD_RLE;
      }
    }
    if (cnt <= MAXPLN_FAST && imgPos >= lineSize) {
      tmpPos = imgPos;
      tmpPosPL = imgPos - lineSize;
      if ((endData = imgPos+MAXPLN_FAST) > imgSize)
        endData = imgSize;
      while (tmpPos < endData && imgDataP[tmpPosPL++] == imgDataP[tmpPos]) tmpPos++;
      tmpCnt = tmpPos-imgPos;
      if (tmpCnt > 1 && tmpCnt >= cnt) {
        cnt = tmpCnt;
        method = METHOD_PLN;
      }
    }
    switch (method) {
      case METHOD_UNC:
        bits_out_ (&bits, 256 + imgDataP[imgPos], 9);
        break;
      case METHOD_PFR:
        if (cnt > 65) {
          bits_out_ (&bits, 256+128+64 + ((cnt-66) & 0x3F), 10);
          length_out_ (&bits, (cnt+62) >> 6);
        }
        else bits_out_ (&bits, 256+128 + (cnt-2), 10);
        break;
      case METHOD_RLE:
        bits_out_ (&bits, imgDataP[imgPos], 2+8);
        switch (resZ) {
          case 2:
            bits_out_ (&bits, imgDataP[imgPos+1], 8);
            break;
          case 3:
            bits_out_ (&bits, imgDataP[imgPos+1], 8);
            bits_out_ (&bits, imgDataP[imgPos+2], 8);
            break;
        }
        length_out_ (&bits, cnt / resZ);
        break;
      case METHOD_PLN:
        bits_out_ (&bits, 128 + (cnt-2), 9);
        break;
    }
    imgPos += cnt;
    if (imgPos >= imgSize) bits_out_ (&bits, 256+128+64, 10); // ****************
    if (bits.bufferP > cmpsMax)
      return (0);
  } while (imgPos < imgSize);
  bits_done_ (&bits);
/*  {
    FILE *fd;
    fd = fopen ("/tmp/x/test", "wb");
    fwrite (cmpsDataP, 1, bits.bufferP - cmpsDataP, fd);
    fclose (fd);
  }*/
  return (bits.bufferP - cmpsDataP);
}


static Uint32  get_length_ (
  HavBits  *bitsP)
{
  Uint32  code;

  while (bitsP->bitsShift >= 8) {
    bitsP->bitsValue = (bitsP->bitsValue << 8) | *(bitsP->bufferP++);
    bitsP->bitsShift -= 8;
  }
  code = bitsP->bitsValue << bitsP->bitsShift;
  if (code & 0x80000000) {
    bitsP->bitsShift += 2;
    return (code >> 30);
  }
  else if (code & 0x40000000) {
    bitsP->bitsShift += 4;
    return (code >> 28);
  }
  else if (code & 0x20000000) {
    bitsP->bitsShift += 6;
    return (code >> 26);
  }
  else if (code & 0x10000000) {
    bitsP->bitsShift += 8;
    return (code >> 24);
  }
  else if (code & 0x8000000) {
    bitsP->bitsShift += 10;
    return (code >> 22);
  }
  else if (code & 0x4000000) {
    bitsP->bitsShift += 12;
    return (code >> 20);
  }
  else if (code & 0x2000000) {
    bitsP->bitsShift += 14;
    return (code >> 18);
  }
  else {
    bitsP->bitsShift += 16;
    return ((code >> 16) + 256);
  }
}


static HavError  hav_decode_fast (
  Uchar   *imgDataP,
	Uchar   *cmpsDataP,
	Uint32  resX,
	Uint32  resY,
	Uint32  bpp)
{
  Uint32   resZ, lineSize, len, code, n;
  Uchar    *imgDataEnd, color1, color2, color3;
  Ushort   *pWord, colorW;
  Uchar    *colorWP = (Uchar *)&colorW;
  HavBits  bits;

  bits.bufferP = cmpsDataP;
  bits.bitsShift = 32;
  resZ = (bpp + 7) / 8;
  lineSize = resX * resZ;
  imgDataEnd = imgDataP + (lineSize * resY);
  do {
    while (bits.bitsShift >= 8) {
      bits.bitsValue = (bits.bitsValue << 8) | *(bits.bufferP++);
      bits.bitsShift -= 8;
    }
    code = bits.bitsValue << bits.bitsShift;
    if (code & 0x80000000) {  // K_METHOD_UNC
      bits.bitsShift += 9;
      *(imgDataP++) = code >> 23;
    }
    else if (code < 0x40000000) {  // K_METHOD_RLE
      switch (resZ) {
        case 1:
          bits.bitsShift += 10;
          len = get_length_ (&bits);
          memset (imgDataP, code >> 22, len);
          imgDataP += len;
          break;
        case 2:
          bits.bitsShift += 18;
          colorWP[0] = code >> 22;
          colorWP[1] = code >> 14;
          len = get_length_ (&bits);
          pWord = (Ushort *) imgDataP;
          for (n = 0; n < len; n++)
            pWord[n] = colorW;
          imgDataP += (len << 1);
          break;
        case 3:
          color1 = code >> 22;
          bits.bitsValue = (bits.bitsValue << 8) | *(bits.bufferP++);
          color3 = code = (bits.bitsValue << (bits.bitsShift+2)) >> 16;
          color2 = code >> 8;
          bits.bitsShift += 18;
          len = get_length_ (&bits);
          for (n = 0; n < len; n++) {
            *(imgDataP++) = color1;
            *(imgDataP++) = color2;
            *(imgDataP++) = color3;
          }
          break;
      }
    }
    else if (code < 0x60000000) {  // K_METHOD_PLN
      bits.bitsShift += 9;
      len = (code >> 23) - 126;
      memmove_safe_ (imgDataP, imgDataP-lineSize, len);
      imgDataP += len;
    }
    else {  // K_METHOD_PRF
      bits.bitsShift += 10;
      code >>= 22;
      if (code < 448)
        imgDataP += (code - 382);
      else
        imgDataP += ((get_length_(&bits) << 6) + (code - (448-66+128)));
    }
  } while (imgDataP < imgDataEnd);
  return (E_HAV_OK);
}


static void  predict_on_ (
  Uchar   *dest,
	Uchar   *src,
	Uint32  pixels,
	Uint32  bpp)
{
  Uchar   color[4] = {0,0,0,0};
  Uint32  c, resZ = (bpp + 7) / 8;

  while (pixels--)
    for (c = 0; c < resZ; c++) {
      *(dest++) = *src - color[c];
      color[c] = *(src++);
    }
}


static void  predict_off_ (
  Uchar   *dest,
	Uchar   *src,
	Uint32  pixels,
	Uint32  bpp)
{
  Uchar   color[4] = {0,0,0,0};
  Uint32  c, resZ = (bpp + 7) / 8;

  while (pixels--)
    for (c = 0; c < resZ; c++)
      *(dest++) = (color[c] += *(src++));
}


typedef struct {
  Ushort  prev, next;
} Qsqr;

typedef struct {
  Ushort    prev, next, len;
  Uint32  imgOfs;
} Qidx;

typedef  Ushort  QsqrFrs[1024];


static Uint32  length_bits_ (
  Uint32  len)
{
  if (len == 0) return (0);
  else if (len <= 3) return (2);
  else if (len <= 7) return (4);
  else if (len <= 15) return (6);
  else if (len <= 31) return (8);
  else if (len <= 63) return (10);
  else if (len <= 127) return (12);
  else if (len <= 255) return (14);
  else return (16);
}


static Uint32  hav_encode_best (
  Uchar   *cmpsDataP,
	Uchar   *imgDataP,
	Uchar   *prevImgDataP,
	Uint32  resX,
	Uint32  resY,
	Uint32  bpp,
  Bool    doPredict)
{
  Bool     usePRFmethod = !!prevImgDataP;
  Int32    imgPos = 0, tmpPos, tmpPosPL, tmpCnt, method, scanLen = 0;
  Int32    endData, resZ, lineSize, imgSize;
  Uchar    *pByte, *cmpsMax, color1, color2, color3;
  Ushort   *pWord;
  QsqrFrs  *sqrFrs;
  Ushort   *idxFrs;
  Qsqr     *sqrTbl;
  Qidx     *idxTbl;
  Int32    posMod, scanPos, fromBlk, toBlk, maxBlk, frsEntry, blk, sqrCenter, sqrLine, cntBits, tmpCntBits;
  Int32    sqrMatch = 0, idxMatch = 0;  // initialize to keep compiler happy
  Int32    sqrImgBigOfs, sqrSize, sqrPos = 0, idxPos = 0, idxMax = 0, sqrFrsSize, n, cnt, sqrImgOfs;
  HavBits  bits;
  Uchar    *iDataP;
                  
  bits_init_ (&bits, cmpsDataP);
  resZ = (bpp + 7) / 8;
  lineSize = resX * resZ;
  imgSize = lineSize * resY;
  cmpsMax = cmpsDataP + (Uint32) (imgSize * 0.95);
  sqrSize = lineSize * SQR_YRES;
  maxBlk = (lineSize + SQR_HALF_XRES - 1) / SQR_HALF_XRES;
  sqrFrsSize = maxBlk * sizeof(QsqrFrs);
  sqrFrs = (QsqrFrs *) malloc (sqrFrsSize);
  idxFrs = (Ushort *) malloc (sizeof(Ushort) * 16384);
  sqrTbl = (Qsqr *) malloc (sizeof(Qsqr) * sqrSize);
  idxTbl = (Qidx *) malloc (sizeof(Qidx) * MAXIDX);
  memset (sqrFrs, 0xFF, sqrFrsSize);
  memset (idxFrs, 0xFF, sizeof(Ushort) * 16384);
  sqrImgBigOfs = - sqrSize;
  if (doPredict) {
    iDataP = (Uchar *) malloc (imgSize);
    predict_on_ (iDataP, imgDataP, resX * resY, bpp);
  }
  else
    iDataP = imgDataP;
  do {
    cnt = MAXUNC;
    cntBits = 0;
    method = METHOD_UNC;
    if (usePRFmethod) {
      tmpPos = imgPos;
      if ((endData = imgPos+MAXPFR) > imgSize)
        endData = imgSize;
      while (tmpPos < endData && iDataP[tmpPos] == prevImgDataP[tmpPos]) tmpPos++;
      tmpCnt = tmpPos-imgPos;
      tmpCntBits = (tmpCnt << 3) - length_bits_ ((tmpCnt+62) >> 6) - 12;
      if (tmpCnt > 1 && tmpCntBits > cntBits) {
        cnt = tmpCnt;
        cntBits = tmpCntBits;
        method = METHOD_PFR;
      }
    }
    if (cnt <= MAXRLE) {
      tmpCnt = (imgSize-imgPos) / resZ;
      if ((endData = MAXRLE) > tmpCnt)
        endData = tmpCnt;
      tmpCnt = 1;
      switch (resZ) {
        case 1:
          pByte = &iDataP[imgPos];
          while (tmpCnt < endData && pByte[0] == pByte[tmpCnt]) tmpCnt++;
          break;
        case 2:
          pWord = (Ushort *) &iDataP[imgPos];
          while (tmpCnt < endData && pWord[0] == pWord[tmpCnt]) tmpCnt++;
          break;
        case 3:
          pByte = &iDataP[imgPos];
          color1 = *(pByte++);
          color2 = *(pByte++);
          color3 = *(pByte++);
          while (tmpCnt < endData && color1 == *(pByte++)
          && color2 == *(pByte++) && color3 == *(pByte++)) tmpCnt++;
          break;
      }
      tmpCntBits = ((tmpCnt-1) << 3) - length_bits_ (tmpCnt) - 4;
      if (tmpCnt > 1 && tmpCntBits > cntBits) {
        cnt = tmpCnt * resZ;
        cntBits = tmpCntBits;
        method = METHOD_RLE;
      }
    }
    if (cnt <= MAXPLN_BEST && imgPos >= lineSize) {
      tmpPos = imgPos;
      tmpPosPL = imgPos - lineSize;
      if ((endData = imgPos+MAXPLN_BEST) > imgSize)
        endData = imgSize;
      while (tmpPos < endData && iDataP[tmpPosPL++] == iDataP[tmpPos]) tmpPos++;
      tmpCnt = tmpPos-imgPos;
      tmpCntBits = (tmpCnt << 3) - 12;
      if (tmpCnt > 1 && tmpCntBits > cntBits) {
        cnt = tmpCnt;
        cntBits = tmpCntBits;
        method = METHOD_PLN;
      }
    }
    if (cnt <= MAXIDX) {
      scanPos = idxFrs[((iDataP[imgPos] << 7) | (iDataP[imgPos+1] & 0x7F)) & 0x3FFF];
      idxMatch = 0xFFFF;
      tmpCnt = cnt-1;
      while (scanPos != 0xFFFF) {
        if (idxTbl[scanPos].len > tmpCnt && imgPos+idxTbl[scanPos].len <= imgSize
        && memcmp (&iDataP[imgPos], &iDataP[idxTbl[scanPos].imgOfs], idxTbl[scanPos].len) == 0) {
          tmpCnt = idxTbl[scanPos].len;
          idxMatch = scanPos;
        }
        scanPos = idxTbl[scanPos].next;
      }
      if (idxMatch != 0xFFFF) {
        tmpCntBits = (tmpCnt << 3) - 12;
        if (tmpCntBits > cntBits) {
          cnt = tmpCnt;
          cntBits = tmpCntBits;
          method = METHOD_IDX;
        }
      }
    }
    if (imgPos > 0 && cnt <= MAXSQR) {
      frsEntry = ((iDataP[imgPos] << 5) | (iDataP[imgPos+1] & 0x1F)) & 0x3FF;
      sqrMatch = 0xFFFF;
      fromBlk = ((imgPos-SQR_HALF_XRES) % lineSize) / SQR_HALF_XRES;
      toBlk = ((imgPos+SQR_HALF_XRES-1) % lineSize) / SQR_HALF_XRES;
      if ((endData = MAXRLE) + imgPos > imgSize)
        endData = imgSize - imgPos;
      if ((tmpCnt = cnt-1) > endData)
        tmpCnt = endData;
      while (fromBlk != toBlk+1) {
        if (fromBlk >= maxBlk) fromBlk = 0;
        scanPos = sqrFrs[fromBlk][frsEntry];
        while (scanPos != 0xFFFF) {
          if (scanPos >= sqrPos) sqrImgOfs = sqrImgBigOfs + scanPos;
          else sqrImgOfs = sqrImgBigOfs + scanPos + sqrSize;
          if (memcmp (&iDataP[imgPos], &iDataP[sqrImgOfs], tmpCnt) == 0) {
            scanLen = tmpCnt;
            while (scanLen < endData
            && iDataP[imgPos+scanLen] == iDataP[sqrImgOfs+scanLen]) scanLen++;
            if (scanLen > tmpCnt) {
              posMod = (sqrImgOfs - ((imgPos-SQR_HALF_XRES) % lineSize)) % lineSize;
              if (posMod >= 0 && posMod < SQR_HALF_XRES*2) {
                tmpCnt = scanLen;
                sqrMatch = sqrImgOfs;
                if (tmpCnt >= endData) break;
              }
            }
          }
          scanPos = sqrTbl[scanPos].next;
        }
        fromBlk++;
      }
      if (sqrMatch != 0xFFFF) {
       tmpCntBits = (tmpCnt << 3) - length_bits_ (tmpCnt) - 12;
       if (tmpCntBits > cntBits) {
         cnt = tmpCnt;
         cntBits = tmpCntBits;  // *****************
         method = METHOD_SQR;
         sqrLine = (imgPos - sqrMatch + lineSize - 1 - SQR_HALF_XRES) / lineSize;
         sqrCenter = imgPos - sqrLine * lineSize;
         if (sqrMatch < sqrCenter)
           sqrMatch = sqrMatch + SQR_HALF_XRES - sqrCenter + sqrLine * SQR_HALF_XRES;
         else
           sqrMatch = sqrMatch - sqrCenter + ((sqrLine-1) * SQR_HALF_XRES) + SQR_HALF_XRES * SQR_YRES;
        }
      }
    }
    switch (method) {
      case METHOD_UNC:
        bits_out_ (&bits, 512+256 + iDataP[imgPos], 10);
        break;
      case METHOD_PFR:  
        if (cnt > 65) {
          bits_out_ (&bits, 256+128+64 + ((cnt-66) & 0x3F), 12);
          length_out_ (&bits, (cnt+62) >> 6);
        }
        else
          bits_out_ (&bits, 256+128 + (cnt-2), 12);
        break;
      case METHOD_RLE:
        bits_out_ (&bits, iDataP[imgPos], 4+8);
        switch (resZ) {
          case 2:
            bits_out_ (&bits, iDataP[imgPos+1], 8);
            break;
          case 3:
            bits_out_ (&bits, iDataP[imgPos+1], 8);
            bits_out_ (&bits, iDataP[imgPos+2], 8);
            break;
        }
        length_out_ (&bits, cnt / resZ);
        break;
      case METHOD_PLN:
        bits_out_ (&bits, 256 + (cnt-2), 12);
        break;
      case METHOD_IDX:
        bits_out_ (&bits, 256+128+128 + idxMatch, 12);
        break;
      case METHOD_SQR:
        bits_out_ (&bits, 256+128+128 + MAXIDX + sqrMatch, 12);
        length_out_ (&bits, cnt);
        if (idxMax == MAXIDX) {
          frsEntry = ((iDataP[idxTbl[idxPos].imgOfs] << 7) | (iDataP[idxTbl[idxPos].imgOfs+1] & 0x7F)) & 0x3FFF;
          if (idxTbl[idxPos].prev == 0xFFFF) idxFrs[frsEntry] = 0xFFFF;
          else idxTbl[idxTbl[idxPos].prev].next = 0xFFFF;
        }
        frsEntry = ((iDataP[imgPos] << 7) | (iDataP[imgPos+1] & 0x7F)) & 0x3FFF;
        if (idxFrs[frsEntry] != 0xFFFF)
          idxTbl[idxFrs[frsEntry]].prev = idxPos;
        idxTbl[idxPos].prev = 0xFFFF;
        idxTbl[idxPos].next = idxFrs[frsEntry];
        idxTbl[idxPos].imgOfs = imgPos;
        idxTbl[idxPos].len = cnt;
        idxFrs[frsEntry] = idxPos;
        if (++idxPos == MAXIDX) idxPos = 0;
        if (idxMax < MAXIDX) idxMax++;
        break;
    }
    if (bits.bufferP > cmpsMax)
      return (0);
    for (n = 1; n <= cnt; n++) {
      if (imgPos >= sqrSize) {
        if (sqrTbl[sqrPos].prev == 0xFFFF) {
          frsEntry = ((iDataP[imgPos-sqrSize] << 5) | (iDataP[imgPos-sqrSize+1] & 0x1F)) & 0x3FF;
          sqrFrs[((imgPos-sqrSize) % lineSize) / SQR_HALF_XRES][frsEntry] = 0xFFFF;
        }
        else sqrTbl[sqrTbl[sqrPos].prev].next = 0xFFFF;
      }
      blk = imgPos % lineSize / SQR_HALF_XRES;
      frsEntry = ((iDataP[imgPos] << 5) | (iDataP[imgPos+1] & 0x1F)) & 0x3FF;
      if (sqrFrs[blk][frsEntry] != 0xFFFF)
        sqrTbl[sqrFrs[blk][frsEntry]].prev = sqrPos;
      sqrTbl[sqrPos].prev = 0xFFFF;
      sqrTbl[sqrPos].next = sqrFrs[blk][frsEntry];
      sqrFrs[blk][frsEntry] = sqrPos;
      if (++sqrPos == sqrSize) {
        sqrPos = 0;
        sqrImgBigOfs += sqrSize;
      }
      imgPos++;
    }
    if (imgPos >= imgSize) bits_out_ (&bits, 256+128+64, 12); // ******************
  } while (imgPos < imgSize);
  bits_done_ (&bits);
  free (sqrFrs);
  free (idxFrs);
  free (sqrTbl);
  free (idxTbl);
  if (doPredict)
    free (iDataP);
/*{
  FILE *fd;
  fd = fopen ("/tmp/x/test", "wb");
  fwrite (cmpsData, 1, cmpsBuf - cmpsData, fd);
  fclose (fd);

}*/
  return (bits.bufferP - cmpsDataP);
}


typedef struct {
  Uchar    *imgPtr;
  Uint32  len;
} QidxDec;


static HavError  hav_decode_best (
  Uchar   *imgDataP,
	Uchar   *cmpsDataP,
	Uint32  resX,
	Uint32  resY,
	Uint32  bpp,
  Bool    doPredict)
{
  Uint32   resZ, lineSize, len, code, n, m, idxPos=0;
  Uchar    *imgDataEnd, color1, color2, color3;
  Ushort   *pWord, colorW;
  Uchar    *colorWP = (Uchar *)&colorW;
  Uint32   sqrTbl[24];
  QidxDec  idxTbl[MAXIDX];
  HavBits  bits;

  bits.bufferP = cmpsDataP;
  bits.bitsShift = 32;
  resZ = (bpp + 7) / 8;
  lineSize = resX * resZ;
  imgDataEnd = imgDataP + (lineSize * resY);
  for (n = 0, m = 64; n < 12; n++, m += lineSize)
    sqrTbl[n] = m;
  for (n = 12, m = lineSize; n < 24; n++, m += lineSize)
    sqrTbl[n] = m;
  do {
    while (bits.bitsShift >= 8) {
      bits.bitsValue = (bits.bitsValue << 8) | *(bits.bufferP++);
      bits.bitsShift -= 8;
    }
    code = bits.bitsValue << bits.bitsShift;
    if (code >= 0xC0000000) {  // K_METHOD_UNC
      bits.bitsShift += 10;
      *(imgDataP++) = code >> 22;
    }
    else if (code < 0x10000000) {  // K_METHOD_RLE
      switch (resZ) {
        case 1:
          bits.bitsShift += 12;
          len = get_length_ (&bits);
          memset (imgDataP, code >> 20, len);
          imgDataP += len;
          break;
        case 2:
          bits.bitsShift += 20;
          colorWP[0] = code >> 20;
          colorWP[1] = code >> 12;
          len = get_length_ (&bits);
          pWord = (Ushort *) imgDataP;
          for (n = 0; n < len; n++) 
            pWord[n] = colorW;
          imgDataP += (len << 1);
          break;
        case 3:
          color1 = code >> 20;
          bits.bitsValue = (bits.bitsValue << 8) | *(bits.bufferP++);
          color3 = code = (bits.bitsValue << (bits.bitsShift+4)) >> 16;
          color2 = code >> 8;
          bits.bitsShift += 20;
          len = get_length_ (&bits);
          for (n = 0; n < len; n++) {
            *(imgDataP++) = color1;
            *(imgDataP++) = color2;
            *(imgDataP++) = color3;
          }
          break;
      }
    }
    else if (code >= 0x60000000) {  // K_METHOD_SQR
      bits.bitsShift += 12;
      len = get_length_ (&bits);
      idxTbl[idxPos].imgPtr = imgDataP;
      idxTbl[idxPos].len = len;
      code = (code >> 20) - 1536;
      memmove_safe_ (imgDataP, imgDataP - sqrTbl[code >> 6] + (code & 0x3F), len);
      imgDataP += len;
      if (++idxPos == MAXIDX) 
        idxPos = 0;
    }
    else if (code >= 0x20000000) {  // K_METHOD_IDX
      bits.bitsShift += 12;
      code = (code >> 20) - 512;
      memmove_safe_ (imgDataP, idxTbl[code].imgPtr, idxTbl[code].len);
      imgDataP += idxTbl[code].len;
    }
    else if (code < 0x18000000) {  // K_METHOD_PLN
      bits.bitsShift += 12;
      len = (code >> 20) - 254;
      memmove_safe_ (imgDataP, imgDataP-lineSize, len);
      imgDataP += len;
    }
    else {  // K_METHOD_PRF
      bits.bitsShift += 12;
      code >>= 20;
      if (code < 448)
        imgDataP += (code - 382);
      else
        imgDataP += ((get_length_ (&bits) << 6) + (code - (448-66+128)));
    }
  } while (imgDataP < imgDataEnd);
  if (doPredict) {
    imgDataP -= (lineSize * resY);
    predict_off_ (imgDataP, imgDataP, resX * resY, bpp);
  }
  return (E_HAV_OK);
}


Uint32  hav_encode (
  Uchar   *cmpsDataP,
  Uchar   *imgDataP, 
  Uchar   *prevImgDataP, 
  Uint32  resX, 
  Uint32  resY, 
  Uint32  bpp, 
  Uint32  codecId,
  Uint32  *trueCodecId)
{
  Uint32  imgSize = resX * resY * ((bpp + 7) / 8);
  Uint32  result = 0;

  switch (codecId) {
    case HAV_VIDEO_CODEC_RAW:
      result = imgSize;
      memmove (cmpsDataP, imgDataP, imgSize);
      break;
    case HAV_VIDEO_CODEC_BEST:
      result = hav_encode_best (cmpsDataP, imgDataP, prevImgDataP, resX, resY, bpp, FALSE);
      break;
    case HAV_VIDEO_CODEC_BEST_PREDICT:
      result = hav_encode_best (cmpsDataP, imgDataP, prevImgDataP, resX, resY, bpp, TRUE);
      break;
    default:  // Fast compression
      result = hav_encode_fast (cmpsDataP, imgDataP, prevImgDataP, resX, resY, bpp);
      break;
  }
  if (result == 0) {
    *trueCodecId = HAV_VIDEO_CODEC_RAW;
    result = imgSize;
    memmove (cmpsDataP, imgDataP, imgSize);
  }
  else
    *trueCodecId = codecId;
  return (result);
}


HavError  hav_decode (
  Uchar*  imgDataP, 
  Uchar*  cmpsDataP,
  Uint32  resX,
  Uint32  resY,
  Uint32  bpp,
  Uint32  codecId)
{
/*  {
    FILE *fd;
    fd = fopen ("/tmp/x/test2", "wb");
    fwrite (cmpsData, 1, imgPacket->dataSize, fd);
    fclose (fd);
  }*/
  switch (codecId) {
    case HAV_VIDEO_CODEC_RAW:
      memmove (imgDataP, cmpsDataP, resX * resY * ((bpp+7) / 8));
      break;
    case HAV_VIDEO_CODEC_FAST:
      hav_decode_fast (imgDataP, cmpsDataP, resX, resY, bpp);
      break;
    case HAV_VIDEO_CODEC_BEST:
      hav_decode_best (imgDataP, cmpsDataP, resX, resY, bpp, FALSE);
      break;
    case HAV_VIDEO_CODEC_BEST_PREDICT:
      hav_decode_best (imgDataP, cmpsDataP, resX, resY, bpp, TRUE);
      break;
  }
  return (E_HAV_OK);
}

